/*
 * @(#)BroadcastConversation.java
 * agentMom package
 * MSE Project: Applying Broadcast/Multicast/Secured Communication to agentMom
 * Department of Computing and Information Sciences
 * Kansas State University
 */
package ksu.cis.mom;

import java.net.*;
import java.util.*;
import java.io.*;

/**
 * <p>The BroadcastConversation class is an abstract class that actually 
 * carries out the broadcast message to all agent under the same local network. 
 * This class provides two main services, read broadcast message and send 
 * broadcast message. It is responsible for passing the messages back and 
 * forth using datagram socket. 
 * <p>There are really two types of conversation classes that can be derived 
 * from the BroadcastConversation class, one for the conversation 
 * initiator and one for the conversation respondent. The basic difference 
 * lies in which constructor is used and the details in the abstract run 
 * method, which must be implemented in the concrete class derived from the 
 * BroadcastConversation class.
 * <p>History<br>
 * 09/12/2003: creation date<br>
 * 11/11/2003: Add more comments suitable for JavaDoc.<br>
 * 12/20/2003: Modify code to make this class extends AgentConversation class 
 * instead of Thread class.
 * @author Scott DeLoach and Chairoj Mekprasertvit
 * @version 1.0, 09/12/2003
 * @since agentMom 2.0
 */

public abstract class BroadcastConversation extends AgentConversation
{
  /**
   * To indicate that the message is a request to start conversation. 
   */
  public static final int START_CONVERSATION = 0;
  
  /**
   * To indicate that the message is not a request to start conversation.
   */
  public static final int CONTENT = 1;
  
  /**
   * Datagram socket used to send and receive broadcast conversation.
   */
  protected DatagramSocket dSocket;
  
  /**
   * Message queue for broadcast conversation.
   * It is a Vector type that stores broadcast message sent to this agent.
   */
  protected Vector broadcast_queue;
  
  /**
   * Name of the broadcast conversation. This name must be unique from others 
   * broadcast conversations within agent since it is used for identifying the 
   * destined conversation when broadcast message is received.
   */
  protected String conversation_name;
  
  /**
   * Default Conversation constructor.
   * @param c - Reference to parent MomObject class.
   */
  public BroadcastConversation(MomObject c)
  {
    super(c);
  }
  
  /**
   * Constructor for conversation initiator
   * @param c - reference to parent class.
   * @param broadcast_address - IP address to send broadcast message.
   * @param port - port for broadcast conversation.
   * @param broadcast_queue - message queue for broadcast conversation.
   */
  public BroadcastConversation(
    MomObject c, 
    InetAddress broadcast_address, 
    int port,
    Vector broadcast_queue)
  {
    super(c, broadcast_address.getHostAddress(), port);
    this.broadcast_queue = broadcast_queue;
    
    try
    {
      this.dSocket = new DatagramSocket();
    }
    catch(IOException e)
    {
      System.out.println(
        " ** Agent "
          + parent.name
          + " failed to connect to broadcast socket"); 
      System.exit(1);
    }
  }
  
  /**
   * Constructor for conversation respondent.
   * @param c - reference to parent class.
   * @param broadcast_address - IP address to send broadcast message.
   * @param port - port for broadcast conversation.
   * @param broadcast_queue - message queue for broadcast conversation.
   * @param m - ksu.cis.mom.Message
   */
  public BroadcastConversation(
    MomObject c, 
    InetAddress broadcast_address, 
    int port,
    Vector broadcast_queue,
    Message m)
  {
    super(c, m);
    this.broadcast_queue = broadcast_queue;
    this.connectionHost = broadcast_address.getHostAddress();
    this.connectionPort = port;

    try
    {
      this.dSocket = new DatagramSocket();
    }
    catch(IOException e)
    {
      System.out.println(
        " ** Agent "
          + parent.name
          + " failed to connect to broadcast socket"); 
      System.exit(1);
    }
  }

  /**
   * <p>This method is responsible for sending the Message m through the
   * datagram socket. It also automatically fill the sender, host, port, 
   * replywith and inreplyto fields in the Message object m using the parent 
   * agents name and port attributes and automatically gets the host name from 
   * the system. 
   * <br>It also converts Message type Object to byte array to be able to send
   * through datagram socket.
   * <br>Note that to send message request a start of new conversation the 
   * correct method is startConversation.
   * @param m - message to send
   * @param replywith - name of the originating conversation.
   * @param inreplyto - name of the destined conversation on the other sides of 
   * agent.
   * @throws IOException
   */
  public void sendMessage(
    Message m, 
    String replywith, 
    String inreplyto) throws IOException
  {
    // Initialize Message
    m.sender = parent.name;
    this.write("Sender is " + parent.getClass().getName());
    
    /* Specify the originating conversation so that the receiving conversation
        knows where to reply to. */
    m.replywith = replywith;
    
    // Specify the destined conversation name.
    m.inreplyto = inreplyto;
    m.port = this.connectionPort;
    
    // To convert object to byte array
    ByteArrayOutputStream bStream = new ByteArrayOutputStream();
    ObjectOutputStream os = 
      new ObjectOutputStream(new BufferedOutputStream(bStream));
    
    this.write("Sending message \"" + m.performative + "\" from " + connectionPort);
    if (m.getContent() != null)
      this.write("Content is: " + m.getContent().getClass().getName());
    os.flush();
    os.writeInt(CONTENT);
    os.writeObject(m);
    os.flush();
    
    // Retrives byte array
    byte[] mesg_in_byte = bStream.toByteArray();
    DatagramPacket p = new DatagramPacket(
      mesg_in_byte, 
      mesg_in_byte.length, 
      InetAddress.getByName(connectionHost), 
      connectionPort);
    dSocket.send(p);
  }

  /**
   * <p>This method is responsible for sending message request a start of new 
   * conversation. The Message m is sent through the datagram socket. 
   * It also automatically fill the sender, host, port, replywith and inreplyto 
   * fields in the Message object m using the parent agents name and port 
   * attributes and automatically gets the host name from the system. 
   * <p>It also converts Message type Object to byte array to be able to send
   * through datagram socket.
   * @param m - message to send
   * @param replywith - name of the originating conversation.
   * @param inreplyto - name of the destined conversation on the other sides of 
   * agent.
   * @throws IOException
   */
  public void startConversation(
      Message m, 
      String replywith, 
      String inreplyto) throws IOException
    {
      // Initialize Message
      m.sender = parent.name;
      this.write("Sender is " + parent.getClass().getName());
      
      /* Specify the originating conversation so that the receiving conversation
         knows where to reply to */
      m.replywith = replywith;
      
      // Specify the destined conversation name.
      m.inreplyto = inreplyto;
      m.port = this.connectionPort;
      
      // To convert object to byte array
      ByteArrayOutputStream bStream = new ByteArrayOutputStream();
      ObjectOutputStream os = 
        new ObjectOutputStream(new BufferedOutputStream(bStream));
      
      this.write("Sending message \"" + m.performative + "\" from " + connectionPort);
      if (m.getContent() != null)
        this.write("Content is: " + m.getContent().getClass().getName());
      os.flush();
      os.writeInt(START_CONVERSATION);
      os.writeObject(m);
      os.flush();
      
      // Retrives byte array
      byte[] mesg_in_byte = bStream.toByteArray();
      DatagramPacket p = new DatagramPacket(
        mesg_in_byte, 
        mesg_in_byte.length, 
        InetAddress.getByName(connectionHost), 
        connectionPort);
      dSocket.send(p);
    }

  /**
   * <p>Method to fetch broadcast message from broadcast message queue
   * destined for this conversation. This method can be considered a blocked 
   * read. This method search message destined for specified conversation name
   * by comparing the String parameter with the inreplyto field in the message.
   * If there is no message for specified conversation, it wait for 1000 
   * miilisecond and try again.
   * @param conversation_name - name of conversation to retrive message
   * @return ksu.cis.mom.Message
   */
  public Message readMessage(String conversation_name)
  {
    Message m = null;
    while(true)
    {
      for (int i = 0; i < broadcast_queue.size(); i++)
      {
        m = (Message) broadcast_queue.elementAt(i);
        if (m.inreplyto.equals(conversation_name))
        {
          broadcast_queue.removeElementAt(i);
          System.out.println("Received " + m.performative + " message");
          return m;
        }
      }
      try
      {
        Thread.sleep(1000);
      }
      catch(InterruptedException ex)
      {
        this.write("Error in BroadcastConversation: " + ex.toString());
      }
    }
  }

  /**
   * <p>Method to fetch broadcast message from broadcast message queue
   * destined for this conversation. This method can be considered a nonblocked 
   * read. This method search message destined for specified conversation name
   * by comparing the String parameter with the inreplyto field in the message.
   * If there is no message for specified conversation, it wait for the amount
   * of specified timeout in miilisecond and then try again. If there is still 
   * no message on the second try, it returns Message with content field "timeout". 
   * @param conversation_name - name of conversation to retrive message
   * @return Message with content field "timeout" if there is no message 
   * destine for specified conversation.
   */
  public Message nonblockedReadMessage(String conversation_name, long timeout)
  {
    int count = 0;
    while(count < 2)
    {
      for (int i = 0; i < broadcast_queue.size(); i++)
      {
        m = (Message) broadcast_queue.elementAt(i);
        if (m.inreplyto.equals(conversation_name))
        {
          broadcast_queue.removeElementAt(i);
          System.out.println("Received " + m.performative + " message");
          return m;
        }
      }
      try
      {
        Thread.sleep(timeout);
        count++;
      }
      catch(InterruptedException ex)
      {
        this.write("Error in BroadcastConversation: " + ex.toString());
      }
    }
    m = new Message();
    m.setContent("timeout");
    return m;
  }

  /**
   * <p>run method for BroadcastConversation class. Each derived conversation 
   * must be run as separate thread and must implement this method regarding to 
   * the type of conversation, initiator or repondent conversation.
   * @see java.lang.Runnable#run()
   */
  public abstract void run();

}// End of BroadcastConversation class
