/*
 * @(#)BroadcastHandler.java 
 * agentMom package
 * MSE Project: Applying Broadcast/Multicast/Secured Communication to agentMom
 * Department of Computing and Information Sciences
 * Kansas State University
 */

package ksu.cis.mom;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.util.Vector;

/**
 * <p>BroadcastHandler is responsible for initializing and starting datagram
 * socket for broadcast conversation. It uses the DatagramSocket class to 
 * send and receive broadcast conversation in form of datagram packet.
 * <p>When an agent is created, it needs to creates a new 
 * broadcast handler thread to be able to receive a start of broadcast 
 * conversation from the other agents. When BroadcastHandler is created, 
 * it creates the DatagramSocket class for broadcast conversation.
 * Below is how an agent can start the BroadcastHandler class.
 * <pre>
 * BroadcastHandler bh = new 
 *   BroadcastHandler(this, this.broadcast_port, broadcast_address);
 * bh.start();
 * </pre>
 * <p>There are two constructors for this class. The first one does not require
 * the maximum datagram received packet size and use the default size 
 * (1024 bytes). The second one allows user to specify the maximum datagram
 * received packet size.
 * <p> Note user need to make sure that the received packet size must be 
 * greater than or equal to the maxSize(1024 by default) variable. Otherwise, 
 * message will be partially received.
 * <p>To create the BroadcastHandler, it requires three/four parameters, a 
 * reference to parent agent object, broadcast port number, broadcast address 
 * and/or maximum datagram received packet size. 
 * When this class is started, the broadcast handler starts a datagram 
 * socket on the indicated port and waits for messages from other agents. All
 * of broadcast messages sent to this agent will be received by this class. 
 * When broadcast handler receives a message, it will check whether the message
 * is a start of new conversation. If it is a start of conversation, the
 * BroadcastHandler simply calls the parent agent's receiveBroadcastConversation
 * method with the datagram socket, received message and broadcast message queue.
 * The agent then verify the received message and start an appropriate 
 * conversation. If the received message is for any broadcast conversation 
 * class, it adds the message to the broadcast message queue. Then the 
 * broadcast conversation class can get the message from this queue later.
 * <p>Note that the broadcast communication has loopback effect(message also 
 * send to itself), but the BroadcastHandler will ignore this message.
 * <p>History:<br>
 * 06/12/2003 - creation date<br>
 * 09/23/2003 - Add more comments suitable for JavaDoc.<br>
 * @author Scott DeLoach and Chairoj Mekprasertvit
 * @version 1.0, 12/23/2003
 * @since agentMom 2.0
 */

public class BroadcastHandler extends Thread
{
  /**
   * Default maximum size of received datagram packet. 
   */
  private int maxSize = 1024;

  /**
   * To indicate that the message is a request to start conversation. 
   */
  public static final int START_CONVERSATION = 0;
  
  /**
   * To indicate that the message is not a request to start conversation.
   */
  public static final int CONTENT = 1;

  /**
   * Reference to parent agent class.
   */
  public Agent parent;

  /**
   * Datagram socket used to send and receive broadcast conversation.
   */
  public DatagramSocket bSocket;

  /**
   * Broadcast port number.
   */
  public int broadcast_port;

  /**
   * Message queue for broadcast conversation.
   * It is a Vector type that stores broadcast message sent to this agent.
   */
  public Vector broadcast_queue;

  /**
   * Flag control receive thread execution. 
   */
  private boolean alive;
  
  /**
   * <p> Constructor for BroadcastHandler class. It is responsible for 
   * initializing necessary variables and creates datagram socket.
   * @param p - reference to parent agent class.
   * @param port - port number for listening for broadcast message.
   * @param b_address - broadcast address. In general, it is in the form 
   * "xxx.xxx.xxx.255" for local broadcast.
   */
  public BroadcastHandler(Agent p, int port, InetAddress b_address)
  {
    // Initialize variables
    // call to superclass(thread class) constructor
    super("BroadcastHandler");
    this.parent = (Agent) p;
    this.broadcast_port = port;
    this.broadcast_queue = p.broadcast_queue;
    this.alive = true;

    // Validate port. Port used must be greater than 1024.
    if (broadcast_port <= 1024)
      throw new IllegalArgumentException(
        " ** Bad port number given to constructor for agent: " + p.name);

    try
    {
      bSocket = new DatagramSocket(broadcast_port);
    }
    catch (IOException e)
    {
      System.out.println(
        " ** Agent "
          + parent.name
          + " failed to connect to broadcast port on port"
          + broadcast_port);
      System.exit(1);
    }
    System.out.println(
          "Agent " + p.name + " BroadcastHandler Started on port " + port);
  }
  
  /**
   * <p> Constructor for BroadcastHandler class. It is responsible for 
   * initializing necessary variables and creates datagram socket. This
   * constructor allows to specify the size of received broadcast packet.
   * @param p - reference to parent agent class.
   * @param port - port number for listening for broadcast message.
   * @param b_address - broadcast address. In general, it is in the form 
   * "xxx.xxx.xxx.255" for local broadcast.
   * @param packetSize - user defined maximum size in byte of received datagram 
   * packet. 
   */
  public BroadcastHandler(
    Agent p, 
    int port,  
    InetAddress b_address, 
    int packetSize)
  {
    // Initialize variables
    // call to superclass(thread class) constructor
    super("BroadcastHandler");
    this.parent = (Agent) p;
    this.broadcast_port = port;
    this.broadcast_queue = p.broadcast_queue;
    this.alive = true;
    this.maxSize = packetSize;

    // Validate port. Port used must be greater than 1024.
    if (broadcast_port <= 1024)
      throw new IllegalArgumentException(
        " ** Bad port number given to constructor for agent: " + p.name);

    try
    {
      bSocket = new DatagramSocket(broadcast_port);
    }
    catch (IOException e)
    {
      System.out.println(
        " ** Agent "
          + parent.name
          + " failed to connect to broadcast port on port"
          + broadcast_port);
      System.exit(1);
    }
  }

  /**
   * <p>Run method for BroadcastHandler. It simply waits for incoming broadcast
   * message from other agents. Then, it checks the message whether it is a 
   * start of new broadcast conversation, or a message for broadcast 
   * conversation class. 
   * <br>Note that this is a blocking read message.
   * @see java.lang.Runnable#run()
   */
  public void run()
  {
    System.out.println("Broadcast Running on " + broadcast_port);
    while (alive)
    {
      try
      {
        // Local variables
        Message m = new Message();
        byte[] b = new byte[maxSize];
        int header = 0;
        DatagramPacket packet = new DatagramPacket(b, b.length);
        // initialize received packet size
        packet.setLength(b.length);
        
        bSocket.receive(packet);
        //System.out.println("from machine" + packet.getAddress());
        
        // convert byte array to object
        ByteArrayInputStream bStream = new ByteArrayInputStream(b);
        ObjectInputStream is =
          new ObjectInputStream(new BufferedInputStream(bStream));
        header = is.readInt();  
        m = (Message) is.readObject();
        this.write("Receive Broadcast Performative " + m.performative);
        
        /* Check header 
         * ignore message if the sender is itself (prevent loopback effect)
         * If it is a message request start of new conversation, call
         * parent agent's receiveBroadcastConversation
         * if it is a message for conversation, put message in broadcast message 
         * queue. 
         */
        if (packet.getAddress().getHostAddress().equals(InetAddress.getLocalHost().getHostAddress()))
        {
          //this.write("Receive message loopback message");
        }
        else if (header == CONTENT)
        {
          this.queueMessage(m);
        }
        else if(header == START_CONVERSATION)
        {
          this.write("Receive a request to start new broadcast conversation.");
          parent.receiveBroadcastConversation(bSocket, m, broadcast_queue);
        }
      }
      catch (IOException e)
      {
        this.write(" ** IOException on port from BroadcastHandler.");
      }
      catch (ClassNotFoundException cnfex)
      {
        this.write(" ** ClassNotFoundException from BroadcastHandler.");
        this.write(" ** Message may not be of Message type.");
      }
    }// End of while loop
  } // End of run method

  /**
   * Method for easily printout information on screen.
   * @param s java.lang.String
   */
  private void write(String s)
  {
    System.out.println(s);
  }

  /**
   * Method for adding new message to the queue.
   * @param m - message to be retrived by broadcast conversation.
   */
  private void queueMessage(Message m)
  {
    this.broadcast_queue.addElement(m);
  }

  /**
   * Set value for isAlive.
   * @param alive - if value is false will make the BroadcastHandler thread to
   * stop the next time this value is checked in the run method.
   */
  public void setAlive(boolean alive)
  {
    this.alive = alive;
  }

} // End of class MulticastHandler
