/*
 * @(#)Component.java
 * agentMom package
 * MSE Project: Applying Broadcast/Multicast/Secured Communication to agentMom
 * Department of Computing and Information Sciences
 * Kansas State University
 */
package ksu.cis.mom;

import java.util.*;
import java.io.*;

/**
 * <p>The Component class is an abstract class that defines the minimum 
 * requirements for a component. This class inherits from <code>MomObject
 * </code> class. It implement the Runnable interface to be able to run as a
 * thread. It require only one parameter, MomObject. MomObject is used to be
 * able to refer to the agent that uses this component.
 * <p>The idea of Component class is to support agent architecture that
 * component performs different tasks. Each component is responsible for 
 * particular tasks. With component support, the design is cleaner and easier by
 * distribute tasks to each component.
 * <p>History<br>
 * 06/01/2003: Add more comments suitable for JavaDoc.
 * 
 * @author Scott DeLoach and Chairoj Mekprasertvit 
 * @version 2.0, 08/16/2003
 * @see ksu.cis.mom.MomObject
 * @see ksu.cis.mom.Agent
 * @since agentMom 1.2
 */

public abstract class Component
  extends MomObject
  implements Runnable, Serializable
{
  /**
   * Message used for internal/external communication.
   */
  protected Message m;
  
  /**
   * Name of the component.
   */
  public String name;
  
  /**
   * Message queue for internal communication. It stores message that will be
   * used for communication with other components.
   */
  private Vector internalMessages = new Vector();
  
  /**
   * Message queue for external communication. It stores message that will be
   * used to make conversation with other agents.
   */
  private Vector externalMessages = new Vector();
  
  /**
   * <p>Default constructor for component class.
   * <br>Creation date: (5/3/2001 9:04:19 AM)
   * @param p - Reference to Momobject class that this component class belong to. 
   */
  public Component(MomObject p)
  {
    this.parent = (Agent) p;
    this.port = parent.port;
  }
  
  /**
   * <p> Return the first message(element) in the Vector externalMessage. If 
   * there is no message, null is return.
   * <br>Creation date: (3/8/01 11:45:19 PM)
   * @return the first message in the Vector externalMessage or null.
   */
  protected Message checkExternal()
  {
    if (externalMessages.size() > 0)
      return (Message) externalMessages.remove(0);
    else
      return null;
  }
  
  /**
   * <p>Return the first message(element) in the Vector internalMessage. If 
   * there is no message, null is return.
   * <br>Creation date: (3/8/01 11:45:19 PM)
   * @return the first message in the Vector internalMessage or null.
   */
  protected Message checkInternal()
  {
    if (internalMessages.size() > 0)
      return (Message) internalMessages.remove(0);
    else
      return null;
  }
  
  /**
   * <p>Add message for external communication to the Vector externalMessage.
   * <br>Creation date: (3/8/01 11:45:19 PM)
   * @param m - message for external communication
   */
  public void enqueueExternal(Message m)
  {
    externalMessages.add(m);
  }
  
  /**
   * <p>Add message for internal component to the Vector internalMessage.
   * <br>Creation date: (3/8/01 11:45:19 PM)
   * @param m - message for internal component.
   */
  public void enqueueInternal(Message m)
  {
    internalMessages.add(m);
  }
  
  /**
   * This method call the agent's sendInternal method and pass the message that
   * will be sent to internal component of the agent.
   * Creation date: (3/8/01 11:45:19 PM)
   * @param m - Message class that will be sent to internal component.
   */
  public void sendInternal(Message m)
  {
    parent.sendInternal(m);
  }
  
  /**
   * <p>Method for printout a String information on terminal screen.
   * @param s String to print out on the terminal screen.
   */
  public void write(String s)
  {
    System.out.println(s);
  }
}
