/*
 * @(#)MessageHandler.java
 * agentMom package
 * MSE Project: Applying Broadcast/Multicast/Secured Communication to agentMom
 * Department of Computing and Information Sciences
 * Kansas State University
 */
package ksu.cis.mom;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.ServerSocket;
import java.net.Socket;

/**
 * <p>The MessageHandler class is used to handle unicast connection from other 
 * agents. It uses the Socket and ServerSocket class to establish a TCP/IP
 * connection stream. When an agent is created, it also needs to creates a new 
 * message handler thread to be able to receive a start of unicast conversation 
 * from the other agents. Below is how an agent can start the MessageHandler 
 * class.
 * <pre>
 * MessageHandler h = new MessageHandler(this.port, this);
 * h.start();
 * </pre>
 * <p>To create the MessageHandler, it requires two parameters, port number and 
 * a reference to parent agent object. When it is started, the message handler 
 * starts a socket server on the indicated port and waits for a connection 
 * from another agent. When a connection is received, the MessageHandler 
 * simply calls the parent agents receiveMessage method with the connection 
 * socket and the input and output streams. The agent then verify the received
 * message and start an appropriate conversation. 
 * <p>History:<br>
 * 09/02/2003: Add more comments suitable for JavaDoc.<br>
 * 12/12/2003: Delete "public String endToken = "END";" since it does not used.
 * @author Scott DeLoach and Chairoj Mekprasertvit
 * @version 2.0, 08/29/2003
 * @since agentMom 1.0
 */

public class MessageHandler extends Thread
{
  /**
   * Port number that the MessageHandler listens for unicast conversation.
   */
  public int portNo;

  /**
   * ServerSocket class used for accepting/constructing unicast conversation. 
   */
  public ServerSocket server;

  /**
   * Reference to parent agent object that use this MessageHandler class.
   */
  public Agent parent;

  /**
   * MessageHandler Constructor
   * @param port - Integer port number.
   * @param p - Reference to parent agent object that use this MessageHandler 
   * class.
   */
  public MessageHandler(int port, Agent p)
  {
    super("MessageHandler"); //call to superclass(thread class) constructor
    parent = (Agent) p;
    if (port <= 1024)
      throw new IllegalArgumentException(
        " ** Bad port number given to Listener constructor for agent "
          + p.name);

    // Try to bind a ServerSocket to the given port
    try
    {
      server = new ServerSocket(port);
    }
    catch (IOException e)
    {
      System.out.println(
        " ** Agent "
          + p.name
          + " Failed to start ServerSocket on port "
          + port);
      System.exit(1);
    }

    // Made the connection, so set the local port number
    portNo = port;
    System.out.println(
      "Agent " + p.name + " MessageHandler Started on port " + port);
  }

  /**
   * <p>Code to perform when this object is garbage collected.
   * Any exception thrown by a finalize method causes the finalization to
   * halt. But otherwise, it is ignored.
   * <p>Insert code to finalize the receiver here.
   * This implementation simply forwards the message to super.
   * User may replace or supplement this.
   * @see java.lang.Object#finalize()
   */
  protected void finalize() throws Throwable
  {
    super.finalize();
  }

  /**
   * <p>Run method for MessageHandler class. It starts by accepting the 
   * connection from other agens, and then initialize input and output streams, 
   * and pass the connection socket, input and output stream to the parent 
   * agent's receiveMessage method. receiveMessage method then verify the input 
   * and start an appropriate conversation.
   * @see java.lang.Runnable#run()
   */
  public void run()
  {
    Socket connection;
    int counter = 1;
    ObjectOutputStream output;
    ObjectInputStream input;
    System.out.println("Unicast Running on " + portNo);

    while (true)
    {
      try
      {
        connection = server.accept();
        System.out.println(
          "Connection "
            + counter++
            + " received from "
            + connection.getInetAddress().getHostName());
        output = new ObjectOutputStream(connection.getOutputStream());
        output.flush();
        input = new ObjectInputStream(connection.getInputStream());
        parent.receiveMessage(connection, input, output);
      }
      catch (InterruptedIOException iioex)
      {
        System.out.println(" ** Socket Server timeout on port " + portNo);
      }
      catch (IOException cnfex)
      {
        System.out.println(" ** IOException on port " + portNo);
      }
    }
  }
}
