/*
 * @(#)SecureMulticastHandler.java 
 * agentMom package
 * MSE Project: Applying Broadcast/Multicast/Secured Communication to agentMom
 * Department of Computing and Information Sciences
 * Kansas State University
 */
 
package ksu.cis.mom;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.SocketTimeoutException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;
import java.util.Vector;
import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;

/**
 * <p>SecureMulticastHandler is responsible for initializing and starting 
 * multicast socket, including joining/leaving multicast group. It also 
 * perform message encryption and decrytion. In generally, it handles
 * secured multicast connection with other agents. It uses the MulticastSocket 
 * class to subscribe to multicast group, and uses symmetric key algorithm to
 * perform encryption and decryption. 
 * <p>When an agent is created, it needs to creates a new secured multicast 
 * handler thread to be able to receive a start of multicast conversation 
 * from other agents.
 * When MulticastHandler class is created, it creates the multicast socket and 
 * join to specified multicast group. Then, it automatically sends a multicast 
 * message to the group indicating that this agent has join the group. When a
 * MulticastHandler receives the message indicating the join, it calls the
 * receiveMulticastJoin method in agent class. 
 * Below is how an agent can start the SecureMessageHandler 
 * class.
 * <pre>
 * SecureMulticastHandler smh = new MessageHandler(this, this.port, 0, secure_multicast_address, key, algorithm);
 * smh.start();
 * </pre>
 * <p>To create the MessageHandler, it requires six/seven parameters, 
 * port number, a reference to parent agent object, time to live of multicast 
 * packet, multicast address, private key for encryption and decryption, an 
 * algorithm to perform encryption and decryption and/or maximum datagram 
 * received packet size.
 * <br>Note that the multicast address is actually a class D IP addresses that
 * are in the range 224.0.0.0 to 239.255.255.255.<br>
 * When this class is started, the secured multicast handler starts a multicast 
 * socket on the indicated port and waits for messages from other agents.
 * When multicast handler receives a message, it will decrypt the message and 
 * check whether the message is a start of new 
 * conversation/join message/leave message/conversation message. If it is a 
 * start of conversation, the SecureMulticastHandler simply calls the parent 
 * agent's receiveSecureMulticastcastConversation method with the multicast socket, 
 * received message, multicast message queue, private key and 
 * the Cipher class.
 * The agent then verify the received message and start an appropriate 
 * conversation. If the received message is for any secured multicast 
 * conversation class, it adds the message to the multicast message queue. 
 * Then the multicast conversation class can get the message from this queue 
 * later.
 * <p>Note that the multicast communication has loopback effect(message also 
 * send to itself), but the SecureMulticastHandler will ignore this message.
 * <br>Note that we use symmetric key algorithm, so it is necessary to keep
 * the key and algorithm consistent among agent in the group.
 * <p>History:<br>
 * 12/01/2003 - creation date<br>
 * 12/27/2003 - Add more comments suitable for JavaDoc.<br>
 * @author Scott DeLoach and Chairoj Mekprasertvit
 * @version 1.0, 12/23/2003
 * @since agentMom 2.0
 */
public class SecureMulticastHandler extends Thread
{
  /**
   * Default maximum size of received multicast packet. 
   */
  public int maxSize = 1024;
  
  /**
   * To indicate that the message is a request to start conversation. 
   */
  public static final int START_CONVERSATION = 0;

  /**
   * To indicate that the message is not a request to start conversation.
   */
  public static final int CONTENT = 1;

  /**
   * Identifies a JOIN multicast PDU
   */
  public static final int JOIN = 2;

  /**
   * Identifies a LEAVE multicast PDU
   */
  public static final int LEAVE = 3;

  /**
   * Reference to parent agent class.
   */
  public Agent parent;

  /**
   * Time-to-live for multicast packets sent out on this MulticastSocket 
   * in order to control the scope of the multicasts.
   */
  public int TimeToLive;

  /**
   * Multicast socket used to send and receive secured multicast message
   */
  public MulticastSocket mSocket;

  /**
   * Secured multicast port
   */
  public int secure_multicast_port;

  /**
   * Message queue for secured multicast conversation.
   * It is a Vector type that stores multicast message sent to this agent.
   */
  public Vector secure_multicast_queue;
  
  /**
   * secured multicast group address
   */
  private InetAddress group;

  /**
   * set this to true to leave the multicast group.
   */
  private boolean leave;
  
  /**
   * Key class to store the secret key for encryption and decryption.
   */
  protected Key key = null;
  
  /**
   * Cipher class to perform encrption and decryption.
   */
  protected Cipher cipher = null;

  /**
   * Default constructor for SecureMulticastHandler. It is responsible for 
   * initializing necessary variables, creates multicast socket and decrypts
   * all received messages.
   * @param p - reference to parent agent class.
   * @param port - port number for listening for secured multicast message.
   * @param ttl - time to live of datagram packet.
   * @param group - multicast address, in the range 224.0.0.0-239.255.255.255
   * @param k - secret key.
   * @param algorithm - Symmetric key algorithm.
   */
  public SecureMulticastHandler(
    Agent p, 
    int port, 
    int ttl, 
    InetAddress group,
    Key k, 
    String algorithm)
  {
    // Initialize variables
    super("SecureMulticastHandler"); //call to superclass(thread class) constructor
    this.parent = p;
    this.secure_multicast_port = port;
    this.secure_multicast_queue = p.secure_multicast_queue;
    this.leave = false;
    this.TimeToLive = ttl;
    this.group = group;
    this.key = k;

    // Validate port. Port used must be greater than 1024.
    if (secure_multicast_port <= 1024)
      throw new IllegalArgumentException(
        " ** Bad port number given to constructor for agent: " + p.name);
   
    // Initialize Cipher class
    try
    {
      this.cipher = Cipher.getInstance(algorithm);
    }
    catch (NoSuchAlgorithmException nsae)
    {
      System.out.println("** Algorithm " + algorithm + " cannot be found.");
      System.exit(1);
    }
    catch (NoSuchPaddingException nspe)
    {
      System.out.println("Padding scheme in " + algorithm + " is not available.");
      System.exit(1);
    }
    System.out.println(
          "Agent " + p.name + " SecureMulticastHandler Started on port " + port);
          
    // Connect MulticastSocket to the given port
    try
    {
      mSocket = new MulticastSocket(secure_multicast_port);
      mSocket.setTimeToLive(TimeToLive);
      //mSocket.setLoopbackMode(true);
      mSocket.joinGroup(group);
      sendJoin();
    }
    catch (IOException e)
    {
      System.out.println(
        " ** Agent "
          + parent.name
          + " failed to connect to multicast port on port"
          + secure_multicast_port); 
      System.exit(1);
    }
  }

  /**
   * Default constructor for SecureMulticastHandler. It is responsible for 
   * initializing necessary variables, creates multicast socket and decrypts
   * all received messages. This constructor allows to specify the size of 
   * received secured multicast packet.
   * @param p - reference to parent agent class.
   * @param port - port number for listening for secured multicast message.
   * @param ttl - time to live of datagram packet.
   * @param group - multicast address, in the range 224.0.0.0-239.255.255.255
   * @param k - secret key.
   * @param algorithm - Symmetric key algorithm.
   */
  public SecureMulticastHandler(
    Agent p, 
    int port, 
    int ttl, 
    InetAddress group, 
    int packetSize,
    Key k, 
    String algorithm)
  {
    // Initialize variables
    super("SecureMulticastHandler"); //call to superclass(thread class) constructor
    this.parent = p;
    this.secure_multicast_port = port;
    this.secure_multicast_queue = p.secure_multicast_queue;
    this.leave = false;
    this.TimeToLive = ttl;
    this.group = group;
    this.maxSize = packetSize;
    this.key = k;

    // Validate port. Port used must be greater than 1024.
    if (secure_multicast_port <= 1024)
      throw new IllegalArgumentException(
        " ** Bad port number given to constructor for agent: " + p.name);

    // Connect MulticastSocket to the given port
    try
    {
      mSocket = new MulticastSocket(secure_multicast_port);
      mSocket.setTimeToLive(TimeToLive);
      mSocket.setLoopbackMode(true);
      mSocket.joinGroup(group);
      sendJoin();
    }
    catch (IOException e)
    {
      System.out.println(
        " ** Agent "
          + parent.name
          + " failed to connect to multicast port on port"
          + secure_multicast_port); 
      System.exit(1);
    }
    
    // Initialize Cipher class
    try
    {
      this.cipher = Cipher.getInstance(algorithm);
    }
    catch (NoSuchAlgorithmException nsae)
    {
      System.out.println("** Algorithm " + algorithm + " cannot be found.");
      System.exit(1);
    }
    catch (NoSuchPaddingException nspe)
    {
      System.out.println("** Padding scheme in " + algorithm + " is not available.");
      System.exit(1);
    }
    System.out.println(
          "Agent " + p.name + " SecureMulticastHandler Started on port " + port);        
  }

  /**
   * <p>Sends a multicast JOIN message to notify other agents.
   * We need to convert Message object to Byte type, and then encrypts the byte
   * information to be able to send encrypted message via multicast socket.
   * @throws IOException
   */
  public void sendJoin()throws IOException
  {
    // Initialize output stream
    ByteArrayOutputStream bStream = new ByteArrayOutputStream();
    ObjectOutputStream os = 
      new ObjectOutputStream(new BufferedOutputStream(bStream));
      
    // Initialize Message
    Message m = new Message();
    m.sender = parent.name;
    m.setContent("Join group");
    os.flush();
    os.writeInt(JOIN);
    os.writeObject(m);
    os.flush();

    // Retrives byte array
    byte[] mesg_in_byte = bStream.toByteArray();
    
    // Encrypt information and send
    try
    {
      System.out.println("** Send Join Group");
      cipher.init(Cipher.ENCRYPT_MODE, key);
      byte[] encrypted_mesg_in_byte =  cipher.doFinal(mesg_in_byte);
      DatagramPacket p = new DatagramPacket(
        encrypted_mesg_in_byte, 
        encrypted_mesg_in_byte.length, 
        this.group, 
        this.secure_multicast_port);
      mSocket.send(p);
      
    }
    catch(InvalidKeyException ike)
    {
      System.out.println("** InvalidKeyException in SecureMulticastHandler");
      System.exit(1);
    }
    catch(BadPaddingException bpe)
    {
      System.out.println("** BadPaddingException in SecureMulticastHandler");
      System.exit(1);
    }
    catch(IllegalBlockSizeException ibse)
    {
      System.out.println("** IllegalBlockSizeException in SecureMulticastHandler");
      System.exit(1);
    }
  }
  
  /**
   * <p>Sends a multicast LEAVE message to notify other agents.
   * We need to convert Message object to Byte type, and then encrypts the byte
   * information to be able to send encrypted message via multicast socket.
   * @throws IOException
   */
  public void sendLeave()throws IOException
  {
    // Initialize output stream
    ByteArrayOutputStream bStream = new ByteArrayOutputStream();
    ObjectOutputStream os = 
      new ObjectOutputStream(new BufferedOutputStream(bStream));
      
    // Initialize Message
    Message m = new Message();
    m.sender = parent.name;
    m.setContent("Leave group");
    os.flush();
    os.writeInt(LEAVE);
    os.writeObject(m);
    os.flush();

    // Retrives byte array
    byte[] mesg_in_byte = bStream.toByteArray();
    
    // Encrypt information and send. Finally leave group.
    try
    {
      cipher.init(Cipher.ENCRYPT_MODE, key);
      byte[] encrypted_mesg_in_byte =  cipher.doFinal(mesg_in_byte);
      DatagramPacket p = new DatagramPacket(
        encrypted_mesg_in_byte, 
        encrypted_mesg_in_byte.length, 
        this.group, 
        this.secure_multicast_port);   
      mSocket.send(p);
      mSocket.leaveGroup(group);
      mSocket.send(p);
      mSocket.leaveGroup(group);
      mSocket.close();
    }
    catch(InvalidKeyException ike)
    {
      System.out.println("** InvalidKeyException in SecureMulticastHandler");
      System.exit(1);
    }
    catch(BadPaddingException bpe)
    {
      System.out.println("** BadPaddingException in SecureMulticastHandler");
      System.exit(1);
    }
    catch(IllegalBlockSizeException ibse)
    {
      System.out.println("** IllegalBlockSizeException in SecureMulticastHandler");
      System.exit(1);
    }
  }
  
  /**
   * Run thread method for SecureMulticastHandler.It simply waits for incoming 
   * encrypted multicast message from other agents. Then, it decrypts message
   * and checks the message whether it is a start of new multicast 
   * conversation, or a message for multicast conversation class. 
   * <p>Note that this is a blocking read message.
   */
  public void run()
  {
    System.out.println("Secured Multicast Running on " + secure_multicast_port);
    boolean alive = true;
    while(alive)
    {
      try
      {
        // Local variables
        Message m = new Message();
        byte[] b = new byte[maxSize];
        int header = 0;
        DatagramPacket packet = new DatagramPacket(b, b.length);
        
        packet.setLength(b.length);
        mSocket.setSoTimeout(1000);
        this.checkLeave();
        mSocket.receive(packet);
        byte[] temp = new byte[packet.getLength()];
        System.out.println("** Received Secured Multicast Message in bit:");
        System.out.println("Bit: ");
        for(int i=0; i<packet.getLength(); i++)
        {
          temp[i] = b[i];
          System.out.print(b[i]+ " ");
        }
        // Decrypt received information
        cipher.init(Cipher.DECRYPT_MODE, key);
        byte decrypted_b[] = cipher.doFinal(temp);
 
        ByteArrayInputStream bStream = new ByteArrayInputStream(decrypted_b);
        ObjectInputStream is = 
          new ObjectInputStream(new BufferedInputStream(bStream));
        header = is.readInt();
        m = (Message) is.readObject();
        //System.out.println("THIS IS CONTENT: " + m.getContent());
        
        // ignore message if the sender is itself (prevent loopback effect)
        if (m.sender.equals(parent.name))
        {
          //this.write("Receive message " + (String) mesg.getContent());
        }
        else if(header == JOIN)
        {
          parent.receiveMulticastJoin(m);
        }
        else if(header == LEAVE)
        {
          parent.receiveMulticastLeave(m);
        }
        else if(header == CONTENT)
        {
          this.queueMessage(m);
        }
        else if(header == START_CONVERSATION)
        {
          parent.receiveSecureMulticastConversation(mSocket, m, secure_multicast_queue, key, cipher.getAlgorithm());
        }
      }
      catch (SocketTimeoutException ste)
      {
        //this.write("** Time out check.");
        this.checkLeave();
      } 
      catch (IOException e)
      {
        this.write(" ** IOException on port from SecureMulticastHandler." + e.toString() + e.getCause() +e.getStackTrace());
        alive = false;
      }
      catch (ClassNotFoundException cnfex)
      {
        this.write(" ** ClassNotFoundException from SecureMulticastHandler.");
        //System.exit(1);
      }
      catch(InvalidKeyException ike)
      {
        System.out.println("** InvalidKeyException in SecureMulticastHandler");
        //System.exit(1);
      }
      catch(BadPaddingException bpe)
      {
        System.out.println("** BadPaddingException in SecureMulticastHandler" + bpe.getMessage() + bpe.getCause());
        //System.exit(1);
      }
      catch(IllegalBlockSizeException ibse)
      {
        System.out.println("** IllegalBlockSizeException in SecureMulticastHandler");
        //System.exit(1);
      }
    }// end while loop
  }// End of run method
  
  /**
   * Method for easily printout information on screen.
   * @param s java.lang.String
   */
  public void write(String s)
  {
    System.out.println(s);
  }
  
  /**
   * Method for adding new message to the queue.
   * @param m - message to be retrived by multicast conversation.
   */
  public void queueMessage(Message m)
  {
    this.secure_multicast_queue.addElement(m);
  }

  /**
   * Set value for leave.
   * @param leave - if value is true, it will make the handler thread to
   * stop the next time this value is checked in the run method.
   */
  public void setLeave(boolean leave)
  {
    this.leave = leave;
  }
  
  /**
   * To check the value of leave attribute.
   */
  public void checkLeave()
  {
    try
    {
      if (leave)
        this.sendLeave();
    }
    catch (IOException ioex)
    {
      this.write("** Can't leave the group. " + ioex.toString());
    }
  }
}
