/*
 * @(#)SecureUnicastConversation.java
 * agentMom package
 * MSE Project: Applying Broadcast/Multicast/Secured Communication to agentMom
 * Department of Computing and Information Sciences
 * Kansas State University
 */
package ksu.cis.mom;

import java.io.*;
import javax.net.ssl.*;

/**
 * <p>The SecureUnicastConversation class is an abstract class that actually 
 * carries out the secured unicast message passing between agents. 
 * This class provides two main services, read secured message and send secured 
 * message.It is responsible for passing the messages back and 
 * forth over the secure socket layers connection. 
 * <p>There are really two types of conversation classes that can be derived 
 * from the SecuredUnicastConversation class, one for the conversation 
 * initiator and one for the conversation respondent. The basic difference 
 * lies in which constructor is used and the details in the abstract run 
 * method, which must be implemented in the concrete class derived from the 
 * SecureUnicastConversation class.
 * <p>Note that java 1.4 is required to be able to use this class.
 * <p>History<br>
 * 09/12/2003: creation date<br>
 * 11/11/2003: Add more comments suitable for JavaDoc.<br>
 * 12/20/2003: Modify code to make this class extends AgentConversation class 
 * instead of Thread class.
 * @author Scott DeLoach and Chairoj Mekprasertvit
 * @version 2.0, 09/23/2003
 * @since agentMom 2.0
 */

public abstract class SecureUnicastConversation extends AgentConversation
{
  /**
   * SSLSocket class is used for connecting secured communication with other 
   * agents.
   */
  public SSLSocket connection;
  
  /**
   * ObjectOutputStream class is used for constructing output stream for sending
   * out the message to other agents.
   */
  protected ObjectInputStream input;
  
  /**
   * ObjectInputStream class is used for constructing input stream for receiving
   * the message from other agents.
   */
  protected ObjectOutputStream output;
  
  /**
   * Default Conversation constructor.
   * @param c - Reference to parent MomObject class.
   */
  public SecureUnicastConversation(MomObject c)
  {
    super(c);
  }
  
  /**
   * Constructor for conversation initiator.
   * @param c - Reference to parent MomObject class.
   * @param hostName - host name that this conversation is connected to.
   * @param su_port - port number of the host that this conversation is 
   * connected to.
   */
  public SecureUnicastConversation(MomObject c, String hostName, int su_port)
  {
    super(c, hostName, su_port);
  }
  
  /**
   * Constructor for conversation respondent.
   * @param sslSocket - javax.net.ssl.SSLSocket
   * @param input - java.io.ObjectInputStream
   * @param output - java.io.ObjectOutputStream
   * @param c - ksu.cis.mom.MomObject
   * @param m - ksu.cis.mom.Message
   */
  public SecureUnicastConversation(
    SSLSocket sslSocket,
    ObjectInputStream input,
    ObjectOutputStream output,
    MomObject c,
    Message m)
  {
    super(c, m);
    this.connection = sslSocket;
    this.input = input;
    this.output = output;
  }
  
  /**
   * <p>This read method allows timeout nonblocking read message. It allows the 
   * read message to "timeout" thus allowing the conversation to check to see 
   * if it has a message without waiting forever. The default value of timeout
   * is 100 milliseconds.
   * @param input java.io.ObjectInputStream
   * @return ksu.cis.mom.Message
   */
  public Message nonblockedReadMessage(ObjectInputStream input)
  {
    Message m = null;
    try
    {
      // set timeout for reading a message
      connection.setSoTimeout(0100);
      m = (Message) input.readObject();
    }
    catch (InterruptedIOException iiex)
    {
      System.out.println(" ** Interrupted IO Exception (timeout) on READ!!!!");
    }
    catch (ClassNotFoundException cnfex)
    {
      System.out.println(" ** ClassNotFoundException ");
    }
    catch (IOException cnfex)
    {
      System.out.println(" ** IOException on port " + connectionPort);
    }
    return m;
  }// end nonblockedReadMessage method
  
  /**
   * <p>This read method is blocking read message. It waits until the message 
   * is arrived.
   * @param input java.io.ObjectInputStream
   * @return ksu.cis.mom.Message
   */
  public Message readMessage(ObjectInputStream input)
  {
    Message m = null;
    try
    {
      m = (Message) input.readObject();
      System.out.println("Received " + m.performative + " message");
    }
    catch (InterruptedIOException iiex)
    {
      System.out.println(" ** Interrupted IO Exception on READ!!!!");
    }
    catch (ClassNotFoundException cnfex)
    {
      System.out.println(" ** ClassNotFoundException ");
    }
    catch (IOException cnfex)
    {
      System.out.println(" ** IOException on port " + connectionPort);
    }
    return m;
  }// end readMessage method
  
  /**
   * <p>run method for conversation class. Each derived conversation must be run 
   * as separate thread and must implement this method regarding to the type of
   * conversation, initiator or repondent conversation.
   * @see java.lang.Runnable#run()
   */
  public abstract void run();
  
  /**
   * <p>This method is responsible for sending the Message m through the
   * ObjectOutputStream output using SSLSocket class. It also automatically 
   * fill the sender, host, and port fields in the Message object m using the 
   * parent agents name and port attributes and automatically gets the host 
   * name from the system. 
   * @param m ksu.cis.mom.Message
   * @param output java.io.ObjectOutputStream
   */
  public void sendMessage(Message m, ObjectOutputStream output)
  {
    try
    {
      m.sender = parent.name;
      System.out.println("Sender is " + parent.getClass().getName());
      m.host = connection.getInetAddress().getHostName();
      System.out.println
        ("Host is " + connection.getInetAddress().getHostName());
      m.port = parent.secure_unicast_port;
      System.out.println
        ("Sending message \"" + m.performative + "\" from " + parent.port);
      if (m.getContent() != null)
        System.out.println
          ("Content is: " + m.getContent().getClass().getName());

      // don't timeout on outputs
      connection.setSoTimeout(0);
      output.writeObject((Object) m);
    }
    catch (IOException e)
    {
      System.out.println(" ** Exception in SecureUnicastConversation: " + e);
    }  
  }// end sendMessage method
}// end class SecureUnicastConversation